from functools import wraps
from itertools import starmap, zip_longest

from ether.ether_types import EthABI
from typing import Any, cast, Callable, Dict, Iterable, List, Tuple, Union


class ABIEncodingError(ValueError):
    ...


def single_item_decoder(
        f: Callable[..., Any]) -> Callable[..., Any]:
    '''Decorator for decoders that use a single item.'''
    @wraps(f)
    def wrapper(b: bytes, *args: int) -> Any:
        if len(b) != 32:
            raise ABIEncodingError(
                f'Expected 32 byte input to decoder {f.__name__}.'
                f'Got {len(b)} bytes')
        return f(b, *args)
    return wrapper


def _inner_type(type_str: str) -> str:
    '''Throws if not an array type'''
    return type_str[:type_str.rindex('[')]


def _array_length(type_str: str) -> int:
    '''Returns the length of a fixed array, or 0'''
    if '[' not in type_str:
        return 0

    num = type_str[type_str.rindex('[') + 1: type_str.rindex(']')]
    if num == '':
        return 0
    length = int(num)
    if length == 1:
        raise NotImplementedError('1-length arrays? Really?')
    return length


def _join_lists(*a: Iterable[int]) -> Iterable[int]:
    '''column addition without carry for lists of integers'''
    return map(sum, zip_longest(*a, fillvalue=0))


def _slots_to_encode(type_str: str, arg: Any) -> Tuple[int, int]:
    '''
    Output represents slots used at each nested headtail layer
    e.g. (3 slots in the head, 15 in the tail)
    '''
    t = type_str

    if type_str == 'string':
        # 1 for the pointer to the body,
        return (1, len(_encode_str(arg)) // 32)

    if type_str == 'bytes':
        # 1 for the pointer to the body,
        return (1, len(_encode_dynamic_bytes(arg)) // 32)

    # dynamic array
    if t[-2:] == '[]':
        inner = _inner_type(type_str)
        tail = _join_lists(*[_slots_to_encode(inner, a) for a in arg])
        return (1, sum(tail) + 1)  # 1 in head, 1 for length in tail

    # fixed array
    if t[-1] == ']':
        inner = _inner_type(type_str)
        arr_len = _array_length(type_str)
        if len(arg) != arr_len:
            raise ABIEncodingError(
                'Declared fixed arr length does not match type. '
                f'Expected {arr_len}, got {len(arg)}')
        usage = list(_join_lists(*[_slots_to_encode(inner, a) for a in arg]))
        return (usage[0], sum(usage[1:]))

    return (1, 0)


def _encode_uint(number: int) -> bytes:
    '''Encode any unsigned integer, throws if >32 bytes'''
    if not isinstance(number, int) or number < 0:
        raise ABIEncodingError(f'Expected uint. Got {type(number)}')
    return number.to_bytes(32, 'big', signed=False)


@single_item_decoder
def _decode_uint(b: bytes) -> int:
    '''Decode an unsigned integer'''
    return int.from_bytes(b, 'big', signed=False)


def _encode_int(number: int) -> bytes:
    '''Encode any signed integer. Throws if 2s complement is >32 bytes'''
    if not isinstance(number, int):
        raise ABIEncodingError(f'Expected int. Got {type(number)}')
    return number.to_bytes(32, 'big', signed=True)


@single_item_decoder
def _decode_int(b: bytes) -> int:
    '''Decode a signed integer'''
    return int.from_bytes(b, 'big', signed=True)


def _encode_fixed_bytes(b: Union[str, bytes]) -> bytes:
    if not isinstance(b, bytes) and not isinstance(b, str):
        raise ABIEncodingError(f'Expected string or bytes. Got {type(b)}')
    if isinstance(b, str):
        decoded = b if b[:2] != '0x' else b[2:]
        payload = bytes.fromhex(decoded)
    else:
        payload = cast(bytes, b)

    padding = b''
    if len(payload) % 32 != 0:
        padding = bytes(32 - (len(payload) % 32))
    return b''.join([payload, padding])


@single_item_decoder
def _decode_fixed_bytes(b: bytes, type_str: str) -> bytes:
    length = int(type_str[5:])
    return b[:length]


def _encode_dynamic_bytes(b: Union[str, bytes]) -> bytes:
    '''Returns JUST the tail portion'''
    length = len(b)
    if isinstance(b, str):
        length = len(b) // 2
        if b[:2] == '0x':
            length -= 1
    payload = _encode_fixed_bytes(b)
    return b''.join([_encode_uint(length), payload])


def _decode_dynamic_bytes(b: bytes) -> bytes:
    # first 32 bytes are the length of the string
    length = _decode_uint(b[:32])
    return b[32: 32 + length]


def _encode_str(arg: str) -> bytes:
    if not isinstance(arg, str):
        raise ABIEncodingError(f'Expected str. Got {type(arg)}')
    payload = arg.encode('utf8')
    return _encode_dynamic_bytes(payload)


def _decode_str(b: bytes) -> str:
    return _decode_dynamic_bytes(b).decode('utf8')


def _encode_address(arg: str) -> bytes:
    # addresses are encoded as uint160s, not bytes20s
    return _encode_uint(int(arg, 16))


@single_item_decoder
def _decode_address(b: bytes) -> str:
    return f'0x{b[-20:].hex()}'


def _encode_dynamic_array(type_str: str, arg: List) -> bytes:
    inner = _inner_type(type_str)
    type_list = [inner for _ in arg]
    return _encode_uint(len(arg)) + encode_many(type_list, arg)


def _decode_dynamic_array(type_str: str, b: bytes) -> List:
    '''Decodes the body (tail portion) of a complex type'''
    inner = _inner_type(type_str)
    length = _decode_uint(b[:32])
    type_list = [inner] * length
    return decode_many(type_list, b[32:])


def _is_dynamic(type_str: str) -> bool:
    return ('[]' in type_str         # dynamic array
            or 'string' in type_str  # strings
            or 'bytes[' in type_str  # bytes
            or type_str == 'bytes')  # bytes


def _is_complex(type_str: str) -> bool:
    if '[][]' in type_str:
        return True  # dynamic array
    # static array of dyanmic types
    return 'string[]' in type_str or 'bytes[]' in type_str


def _encode_fixed_array(type_str: str, arg: List) -> bytes:
    '''Head-only lists, no strings, or dynamic entries'''

    return b''.join([encode(_inner_type(type_str), a)[0] for a in arg])


def _decode_fixed_array(type_str: str, b: bytes) -> List:
    '''Head-only lists, no strings, or dynamic entries'''
    if _is_complex(type_str):
        raise ABIEncodingError('Complex type in fixed array decoder')
    length = _array_length(type_str)
    inner = _inner_type(type_str)
    item_len = 32 * _slots_to_encode(inner, range(length))[0]

    items = [b[i:i + item_len] for i in range(0, len(b), item_len)]

    decoded = [decode(inner, item) for item in items]

    return decoded


def encode(
        type_str: str,
        arg: Union[int, bool, str, bytes, List]) -> Tuple[bytes, bytes]:
    '''
    Encode a single item as a bytestring
    Raises an ABIEncodingError if the type does not match the argument
    If the type is dynamic, returns just the tail portion.
    '''
    if type_str == 'address':
        return _encode_address(cast(str, arg)), b''

    elif type_str == 'string':
        return b'', _encode_str(cast(str, arg))

    elif type_str == 'bytes':
        return b'', _encode_dynamic_bytes(cast(bytes, arg))

    elif type_str[-2:] == '[]':
        return b'', _encode_dynamic_array(type_str, cast(List, arg))

    elif '[' in type_str:
        return _encode_fixed_array(type_str, cast(list, arg)), b''

    elif 'bytes' in type_str and int(type_str[5:]) <= 32:
        a = cast(bytes, arg)
        if len(a) > 32:
            raise ABIEncodingError(f'Too long for fixed encoding: {len(a)}B')
        return _encode_fixed_bytes(a), b''

    elif type_str[:4] == 'uint':
        return _encode_uint(cast(int, arg)), b''

    elif type_str[:3] == 'int':
        return _encode_int(cast(int, arg)), b''

    elif type_str == 'bool':
        if not isinstance(arg, bool):
            raise ABIEncodingError(f'Expected bool. Got {type(arg)}')
        b = 1 if arg else 0
        return _encode_uint(b), b''

    elif 'fixed' in type_str:
        raise NotImplementedError('No fixed point numbers. Who even are you?')

    raise ValueError(f'Unknown type: {type_str}')


def _encode_offset(head_size: int, tail_pos: int) -> bytes:
    '''Encode the offset'''
    return _encode_uint((tail_pos + head_size) * 32)


def encode_many(type_list: List[str], arg_list: List[Any]) -> bytes:
    '''Encode many arguments into a single blob.'''
    head: List[bytes] = []
    tail: List[bytes] = []

    slot_usage = list(starmap(_slots_to_encode, zip(type_list, arg_list)))

    # will need to be redone for dynamic types
    head_size = sum([s[0] for s in slot_usage])
    head_pos = 0
    for (i, t, a) in zip(range(len(type_list)), type_list, arg_list):
        encoded_head, encoded_tail = encode(t, a)

        if encoded_head == b'':
            tail_pos = sum(s[1] for s in slot_usage[:i])
            offset = _encode_offset(head_size, tail_pos)
            encoded_head = offset

        head_pos += 1
        head.append(encoded_head)

        if encoded_tail != b'':
            tail.append(encoded_tail)

    return b''.join([*head, *tail])


def encode_tuple(type_tuple: str, args: List) -> bytes:
    '''Encode using an abi-format type-tuple e.g. "(bytes,int,address[])"'''
    # strip () and pass through
    type_tuple = type_tuple[1:-1]
    return encode_many(type_tuple.split(','), args)


def decode(type_str: str, b: bytes) -> Any:
    '''Decode an item from a bytestring'''
    if type_str[-2:] == '[]':
        return _decode_dynamic_array(type_str, b)
    elif '[' in type_str:
        return _decode_fixed_array(type_str, b)
    elif type_str == 'address':
        return _decode_address(b[:32])
    elif 'string' in type_str:
        return _decode_str(b)
    elif 'bytes' in type_str:
        if type_str == 'bytes':
            return _decode_dynamic_bytes(b)
        return _decode_fixed_bytes(b[:32], type_str)
    elif 'uint' in type_str:
        return _decode_uint(b[:32])
    elif 'int' in type_str:
        return _decode_int(b[:32])
    elif type_str == 'bool':
        return b[31:32] == b'\x01'
    elif 'fixed' in type_str:
        raise NotImplementedError('No fixed point numbers. Who even are you?')
    raise ValueError(f'Could not deserialize {type_str} with value {b.hex()}')


def decode_many(type_list: List[str], b: bytes) -> List[Any]:
    if len(b) % 32 != 0:
        raise ValueError('Invalid ABI blob. Expected multiple of 32 bytes')
    # arrays use extra slots depending on their length
    array_head_slots = sum([max(0, _array_length(t) - 1) for t in type_list])
    head_size = len(type_list) + array_head_slots

    # break up the whole thing into its 32-byte slots
    slots = [b[i:i + 32] for i in range(0, len(b), 32)]

    decoded = []
    head_pos = 0
    type_pos = 0
    while head_pos < head_size:
        outer = type_list[type_pos]

        # how many head items are in this?
        array_length = _array_length(outer)
        head_items = max(1, array_length)

        dec = []
        for i in range(head_items):
            h = slots[head_pos]
            t = outer if array_length == 0 else _inner_type(outer)
            if not _is_dynamic(t):
                value = decode(t, h)
            else:
                tail_loc = _decode_uint(h) // 32
                value = decode(t, b''.join(slots[tail_loc:]))
            dec.append(value)
            head_pos += 1
        type_pos += 1

        if array_length != 0:
            decoded.append(dec)
        else:
            decoded.extend(dec)
    return decoded


def decode_tuple(type_tuple: str, b: bytes) -> List:
    '''Decode an abi-format type-tuple e.g. "(bytes,int,address[])"'''
    # strip () and pass through
    type_tuple = type_tuple[1:-1]
    return decode_many(type_tuple.split(','), b)


def find(name: str, interface: EthABI) -> List[Dict[str, Any]]:
    return [e for e in interface if 'name' in e and e['name'] == name]


def _make_type_tuple(f: Dict[str, Any]) -> str:
    '''
    makes a comma-delimted type tuple
    for use in the signature and encoding
    e.g. '(bytes,bytes,bytes)'
    '''
    inputs: List[Dict[str, Any]] = f['inputs']
    types = ','.join(t['type'] for t in inputs)
    return f'({types})'


def make_signature(f: Dict[str, Any]) -> str:
    '''
    Parses a function or event ABI into a signture

    Args:
        f (dict): the function or event ABI
    Returns:
        (str): the signature
    '''
    type_tuple = _make_type_tuple(f)
    return f'{f["name"]}{type_tuple}'



import unittest

from ether import events

from ether.tests import helpers


class TestABI(unittest.TestCase):

    def test_decode_event(self):
        for i in range(len(helpers.weth_transfers)):
            decoded = events.decode_event(
                helpers.weth_transfers[i],
                helpers.weth_json)
            self.assertEqual(decoded, helpers.weth_parsed[i])

    def test_match_topic0_to_event(self):
        with self.assertRaises(ValueError) as context:
            events._match_topic0_to_event(
                'nope',
                [e for e in helpers.weth_json if e['type'] == 'event'])
        self.assertIn('Topic not found', str(context.exception))



from Cryptodome.Hash import keccak
import warnings

from typing import Callable, cast
from ether.ether_types import EthSig

# suppress load warning
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from eth_keys import datatypes as eth_ecdsa


def keccak256(msg: bytes) -> bytes:
    '''
    Does solidity's dumb keccak

    Args:
        msg (bytes): the message to hash
    Returns:
        (bytes): the keccak256 digest
    '''
    keccak_hash = keccak.new(digest_bits=256)
    keccak_hash.update(msg)
    return keccak_hash.digest()


def pow_mod(x: int, y: int, z: int) -> int:
    '''
    int, int, int (or float)
    returns (x^y)mod z
    '''
    number = 1
    while y:
        if y & 1:
            number = number * x % z
        y >>= 1
        x = x * x % z
    return number


def uncompress_pubkey(pubkey: bytes) -> bytes:
    '''
    takes a compressed pubkey, returns the uncompressed pubkey (64 bytes)
    '''
    p = 0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f
    parity = pubkey[0] - 2
    x = int.from_bytes(pubkey[1:], 'big')
    a = (pow_mod(x, 3, p) + 7) % p
    y = pow_mod(a, (p + 1) // 4, p)
    if y % 2 != parity:
        y = -y % p
    return (x.to_bytes(32, 'big')) + (y.to_bytes(32, 'big'))


def compress_pubkey(pubkey: bytes) -> bytes:
    '''Take a an uncompressed pubkey, return the compressed representation'''
    pub = pubkey[1:] if len(pubkey) == 65 else pubkey
    parity = (pub[-1] & 1) + 2
    compressed = bytes([parity]) + pub[:32]
    return compressed


def priv_to_pub(privkey: bytes) -> bytes:
    '''Return the pubkey that corresponds to a private key'''
    priv = eth_ecdsa.PrivateKey(privkey)
    pub = eth_ecdsa.PublicKey.from_private(private_key=priv)
    return cast(bytes, pub.to_bytes())


def pub_to_addr(pubkey: bytes) -> str:
    '''Eth addr is last 20 bytes of keccak256 of pubkey'''
    return f'0x{keccak256(pubkey)[-20:].hex()}'


def priv_to_addr(privkey: bytes) -> str:
    '''Make address from privkey'''
    return pub_to_addr(priv_to_pub(privkey))


def recover_pubkey(signature: EthSig, digest: bytes) -> bytes:
    '''Recovers the public key from a signature and message digest'''
    # bullshit in the underlying eth library
    # needs to be 0 if v is odd, 1 if v is even
    normalized_v = (signature[0] + 1) % 2
    normalized_sig = (normalized_v, signature[1], signature[2])

    sig = eth_ecdsa.Signature(vrs=normalized_sig)
    pub = sig.recover_public_key_from_msg_hash(digest)
    return cast(bytes, pub.to_bytes())


def recover_address(signature: EthSig, digest: bytes) -> str:
    return pub_to_addr(recover_pubkey(signature, digest))


def _der_minimal_int(number: int) -> bytes:
    if number < 0:
        raise ValueError('Negative number in signature')
    return number.to_bytes((number.bit_length() + 7) // 8, 'big')


def sig_to_der(signature: EthSig) -> bytes:
    '''
    0x30|b1|0x02|b2|r|0x02|b3|s
    b1 = Length of remaining data
    b2 = Length of r
    b3 = Length of s
    '''
    r = _der_minimal_int(signature[1])
    s = _der_minimal_int(signature[2])

    enc_r = bytes([0x02, len(r)]) + r
    enc_s = bytes([0x02, len(s)]) + s

    der = bytes([0x30, len(enc_r) + len(enc_s)]) + enc_r + enc_s
    return der


def sign_hash(digest: bytes, privkey: bytes) -> EthSig:
    '''Sign a digest'''
    priv = eth_ecdsa.PrivateKey(privkey)
    sig = priv.sign_msg_hash(digest)
    return cast(EthSig, sig.vrs)


def sign(
        message: bytes,
        privkey: bytes,
        algo: Callable[[bytes], bytes] = keccak256) -> EthSig:
    '''
    Gets a signature on a message digest of a message
    '''
    return sign_hash(algo(message), privkey)


def sign_message(
        message: bytes,
        privkey: bytes,
        algo: Callable[[bytes], bytes] = keccak256) -> EthSig:
    '''Sign a message using the ethereum signed message format'''
    prefixed = b''.join([b'\x19Ethereum Signed Message:\n', message])
    return sign(prefixed, privkey)



from typing import List, Tuple, Union

LIST = False
BYTES = True

# defined by RLP standard
BYTES_OFFSET = 0x80
LONG_BYTES_OFFSET = BYTES_OFFSET + 55  # 0xb7
ARRAY_OFFSET = 0xc0
LONG_ARRAY_OFFSET = ARRAY_OFFSET + 55  # 0xf7


class RLPError(ValueError):
    ...


def i2be_rlp_padded(
        number: int,
        length: int = 0,
        signed: bool = False) -> bytes:
    '''
    Convert int to big endian (b.e.) bytes

    Args:
        number: int value to convert to bytes in BE format
    Returns:
        bytes in BE format
    '''
    if length == 0:
        sign_bit = 1 if signed else 0
        length = (number.bit_length() + 7 + sign_bit) // 8
    if number == 0 and length == 0:
        return b''
    return number.to_bytes(length, 'big', signed=signed)


def be2i_rlp(b: bytes, signed: bool = False) -> int:
    if b == b'':
        return 0
    return int.from_bytes(b, 'big', signed=signed)


def encode(item: Union[List, bytes]) -> bytes:
    if isinstance(item, bytes):
        if len(item) == 1 and item[0] < BYTES_OFFSET:
            return item
        else:
            return _encode_length(len(item), BYTES_OFFSET) + item
    elif isinstance(item, list):
        output = b''.join(encode(i) for i in item)
        return _encode_length(len(output), ARRAY_OFFSET) + output


def _encode_length(length: int, offset: int) -> bytes:
    '''Encode a length '''
    if length <= 55:
        return bytes([offset + length])

    if length >= 256 ** 8:
        raise RLPError('Bytestring is too long to encode')

    enc_len = i2be_rlp_padded(number=length)
    tag = bytes([offset + 55 + len(enc_len)])
    return b''.join([tag, enc_len])


def decode_list(raw: bytes) -> List:
    output = []
    remaining = raw[:]
    while len(remaining) > 0:
        (offset, data_len, type) = _decode_length(remaining)
        try:
            next_item = remaining[:offset + data_len]
            output.append(decode(next_item))
        except IndexError as e:
            raise RLPError('Malformatted bytestring. Overran input.') from e
        remaining = remaining[offset + data_len:]
    return output


def decode(raw: bytes) -> Union[List, bytes]:
    if len(raw) == 0:
        return b''

    (offset, data_len, type) = _decode_length(raw)

    try:
        next_item = raw[offset: offset + data_len]
    except IndexError as e:
        raise RLPError('Malformatted bytestring. Overran input') from e

    if type == BYTES:
        return next_item

    return decode_list(next_item)


def _decode_length(raw: bytes) -> Tuple[int, int, bool]:
    try:
        tag = raw[0]
    except IndexError as e:
        raise RLPError('Malformatted bytestring. Null.') from e

    # single byte
    if tag < BYTES_OFFSET:
        return (0, 1, BYTES)

    # short bytestring
    if tag <= LONG_BYTES_OFFSET:
        bytes_len = tag - BYTES_OFFSET
        return (1, bytes_len, BYTES)

    # long bytestring
    if tag <= LONG_BYTES_OFFSET + 7:
        enc_len_bytes = tag - LONG_BYTES_OFFSET
        enc_len = raw[1: 1 + enc_len_bytes]
        bytes_len = int.from_bytes(enc_len, 'big')
        return (1 + enc_len_bytes, bytes_len, BYTES)

    # short list
    if tag <= LONG_ARRAY_OFFSET:
        list_len = tag - ARRAY_OFFSET
        return (1, list_len, LIST)

    # long list
    enc_len_bytes = tag - LONG_ARRAY_OFFSET
    enc_len = raw[1: 1 + enc_len_bytes]
    list_len = int.from_bytes(enc_len, 'big')
    return (1 + enc_len_bytes, list_len, LIST)



import abc
from ether import crypto, rlp

from typing import Any, cast, List, Optional
from ether.ether_types import CeloTxDict, EthSig, EthTxDict


class Immutable:
    '''
    A simple inheritable class that prevents accidental changes to an object.
    '''
    __immutable: bool = False

    def _make_immutable(self) -> None:
        '''
        Prevents any future changes to the object
        '''
        self.__immutable = True

    def __setattr__(self, key: str, value: Any) -> None:
        '''Override __setattr__ to check if immutable is set before setting'''
        if self.__immutable:
            raise TypeError("%r cannot be written to." % self)
        object.__setattr__(self, key, value)


class EthTx(abc.ABC, Immutable):
    '''
    Abstract base class for Ethereum-like transactions. Contains information
    shared by signed and unsigned txns across networks.

    Args:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata

    Attributes:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
    '''
    nonce: int
    gasPrice: int
    gas: int
    to: str
    value: int
    data: bytes

    def __ne__(self, other):  # type: ignore
        '''Define != operator'''
        return not self == other

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, raw: bytes) -> 'EthTx':
        '''Deserialize a transaction from raw bytes'''
        ...

    @classmethod
    def deserialize_hex(cls, raw: str) -> 'EthTx':
        '''Serialize a transaction from hex.'''
        r = raw
        if r[:2] == '0x':
            r = r[2:]
        return cls.deserialize(bytes.fromhex(r))

    @abc.abstractmethod
    def serialize(self) -> bytes:
        '''Serialize the transaction to raw bytes.'''
        ...

    def serialize_hex(self) -> str:
        '''Serialize the transaction as hex.'''
        return f'0x{self.serialize().hex()}'

    @abc.abstractmethod
    def sighash(self) -> bytes:
        '''
        Calculate the signature hash of the transaction. This is the digest
        that is signed by the sender.
        '''
        ...

    def to_json_dict(self) -> EthTxDict:
        '''
        Convert the transaction into a simple json dict suitable for RPC usage.
        '''
        return EthTxDict(
            nonce=hex(self.nonce),
            gasPrice=hex(self.gasPrice),
            gas=hex(self.gas),
            to=self.to,
            value=hex(self.value),
            data=f'0x{self.data.hex()}'
        )


class CeloTx(EthTx):
    '''
    Shared base class for Celo transactions. Contains Celo-specific information
    used in Signed and Unsigned variants.

    Attributes:
        gasCurrency: the address of the token to be used as gas
        gasFeeRecipient: the address of the recipient of gas fees
    '''
    gasCurrency: Optional[str]
    gasFeeRecipient: Optional[str]

    def to_json_dict(self) -> CeloTxDict:
        '''
        Convert the transaction into a simple json dict suitable for RPC usage.
        '''
        return CeloTxDict(
            nonce=hex(self.nonce),
            gasPrice=hex(self.gasPrice),
            gas=hex(self.gas),
            gasCurrency=self.gasCurrency,
            gasFeeRecipient=self.gasFeeRecipient,
            to=self.to,
            value=hex(self.value),
            data=f'0x{self.data.hex()}'
        )


class UnsignedEthTx(EthTx):
    '''
    An unsigned Ethereum transaction intended for a specific chain.

    Args:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        chainId: the EIP-155 chainId identifying the intended destination chain

    Attributes:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        chainId: the EIP-155 chainId identifying the intended destination chain
    '''
    chainId: int

    def __init__(
            self,
            nonce: int,
            gasPrice: int,
            gas: int,
            to: str,
            value: int,
            data: bytes,
            chainId: int):
        self.nonce = nonce
        self.gasPrice = gasPrice
        self.gas = gas
        self.to = to
        self.value = value
        self.data = data
        self.chainId = chainId
        self._make_immutable()

    def __eq__(self, other):  # type: ignore
        '''Define == operator'''
        if not isinstance(other, type(self)):
            raise TypeError('Unsupported types for equality:'
                            f'{type(self)} and {type(other)}')
        return (
            self.nonce == other.nonce
            and self.gasPrice == other.gasPrice
            and self.gas == other.gas
            and self.to == other.to
            and self.value == other.value
            and self.data == other.data
            and self.chainId == other.chainId
        )

    @classmethod
    def deserialize(cls, raw: bytes) -> 'UnsignedEthTx':
        '''Deserialize a transaction from raw bytes'''
        fake_signed = SignedEthTx.deserialize(raw)
        return fake_signed.as_unsigned()

    def get_signature(self, key: bytes) -> EthSig:
        '''Get a signature on a tx under a certain key'''
        # NB: the underlying signing library sets v to opposite parity
        #     0 = odd
        #     1 = even
        (v, r, s) = crypto.sign_hash(self.sighash(), key)
        v += self.chainId * 2 + 35
        return (v, r, s)

    def serialize(self) -> bytes:
        '''Serialize the transaction to raw bytes.'''
        return self._with_null_signature().serialize()

    def sighash(self) -> bytes:
        '''
        Calculate the signature hash of the transaction. This is the digest
        that is signed by the sender.
        '''
        stripped = self._with_null_signature()
        return crypto.keccak256(stripped.serialize())

    def sign(self, key: bytes) -> 'SignedEthTx':
        '''Return a copy of the transaction signed by the key'''
        signature = self.get_signature(key)
        return SignedEthTx(
            self.nonce,
            self.gasPrice,
            self.gas,
            self.to,
            self.value,
            self.data,
            *signature)

    def with_chainId(self, id: int) -> 'UnsignedEthTx':
        return UnsignedEthTx(
            self.nonce,
            self.gasPrice,
            self.gas,
            self.to,
            self.value,
            self.data,
            id)

    def _with_null_signature(self) -> 'SignedEthTx':
        return SignedEthTx(
            self.nonce,
            self.gasPrice,
            self.gas,
            self.to,
            self.value,
            self.data,
            self.chainId, 0, 0)


class UnsignedCeloTx(CeloTx, UnsignedEthTx):
    '''
    An unsigned Celo transaction intended for a specific chain.

    Args:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        gasCurrency: the address of the token to be used as gas
        gasFeeRecipient: the address of the recipient of gas fees
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        chainId: the EIP-155 chainId identifying the intended destination chain

    Attributes:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        gasCurrency: the address of the token to be used as gas
        gasFeeRecipient: the address of the recipient of gas fees
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        chainId: the EIP-155 chainId identifying the intended destination chain
    '''
    chainId: int

    def __init__(
            self,
            nonce: int,
            gasPrice: int,
            gas: int,
            gasCurrency: Optional[str],
            gasFeeRecipient: Optional[str],
            to: str,
            value: int,
            data: bytes,
            chainId: int):
        self.nonce = nonce
        self.gasPrice = gasPrice
        self.gas = gas
        self.to = to
        self.value = value
        self.data = data
        self.chainId = chainId
        self.gasCurrency = gasCurrency
        self.gasFeeRecipient = gasFeeRecipient
        self._make_immutable()

    def __eq__(self, other):  # type: ignore
        '''Define == operator'''
        if not isinstance(other, type(self)):
            raise TypeError('Unsupported types for equality:'
                            f'{type(self)} and {type(other)}')
        return (
            self.nonce == other.nonce
            and self.gasPrice == other.gasPrice
            and self.gas == other.gas
            and self.gasCurrency == other.gasCurrency
            and self.gasFeeRecipient == other.gasFeeRecipient
            and self.to == other.to
            and self.value == other.value
            and self.data == other.data
            and self.chainId == other.chainId
        )

    @classmethod
    def deserialize(cls, raw: bytes) -> 'UnsignedCeloTx':
        '''Deserialize a transaction from raw bytes'''
        fake_signed = SignedCeloTx.deserialize(raw)
        return fake_signed.as_unsigned()

    def serialize(self) -> bytes:
        '''Serialize the transaction to raw bytes.'''
        return self._with_null_signature().serialize()

    def sign(self, key: bytes) -> 'SignedCeloTx':
        '''Return a copy of the transaction signed by the key'''
        signature = self.get_signature(key)
        return SignedCeloTx(
            self.nonce,
            self.gasPrice,
            self.gas,
            self.gasCurrency,
            self.gasFeeRecipient,
            self.to,
            self.value,
            self.data,
            *signature)

    def _with_null_signature(self) -> 'SignedCeloTx':
        '''
        Return a copy of the txn with a null signature. Used for serializing
        unsigned transactions
        '''
        return SignedCeloTx(
            self.nonce,
            self.gasPrice,
            self.gas,
            self.gasCurrency,
            self.gasFeeRecipient,
            self.to,
            self.value,
            self.data,
            self.chainId, 0, 0)


class SignedEthTx(EthTx):
    '''
    A signed Ethereum transaction.

    Args:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        v: EIP-155 signature v value
        r: signature r value
        s: signature s value

    Attributes:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        v: EIP-155 signature v value
        r: signature r value
        s: signature s value
    '''
    v: int
    r: int
    s: int
    tx_id: bytes

    def __init__(
            self,
            nonce: int,
            gasPrice: int,
            gas: int,
            to: str,
            value: int,
            data: bytes,
            v: int,
            r: int,
            s: int):
        self.nonce = nonce
        self.gasPrice = gasPrice
        self.gas = gas
        self.to = to
        self.value = value
        self.data = data
        self.v = v
        self.r = r
        self.s = s
        self._set_tx_id()
        self._make_immutable()

    def __eq__(self, other):  # type: ignore
        '''Define == operator'''
        if not isinstance(other, type(self)):
            raise TypeError('Unsupported types for equality:'
                            f'{type(self)} and {type(other)}')
        return self.tx_id == other.tx_id

    def as_unsigned(self) -> 'UnsignedEthTx':
        '''Return an unsigned copy of the transaction'''
        # if r == 0 it's an unsigned tx anyway
        chainId = self.v if self.r == 0 \
            else (self.v - (35 if self.v % 2 else 36)) // 2
        return UnsignedEthTx(
            nonce=self.nonce,
            gasPrice=self.gasPrice,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            chainId=chainId)

    @classmethod
    def deserialize(cls, raw: bytes) -> 'SignedEthTx':
        '''Deserialize a transaction from raw bytes'''
        raws = cast(List[bytes], rlp.decode(raw))
        nonce = rlp.be2i_rlp(raws[0])
        gasPrice = rlp.be2i_rlp(raws[1])
        gas = rlp.be2i_rlp(raws[2])
        to = f'0x{raws[3].hex()}'
        value = rlp.be2i_rlp(raws[4])
        data = raws[5]
        (v, r, s) = [rlp.be2i_rlp(i) for i in raws[6:]]

        return SignedEthTx(
            nonce,
            gasPrice,
            gas,
            to,
            value,
            data,
            v, r, s)

    def recover_sender(self) -> str:
        '''Recover the address corresponding to the key that signed this tx'''
        pubkey = crypto.recover_pubkey(
            (self.v, self.r, self.s),
            self.sighash())
        return crypto.pub_to_addr(pubkey)

    def serialize(self) -> bytes:
        '''Serialize the transaction to raw bytes.'''
        to_serialize: List[bytes] = [
            rlp.i2be_rlp_padded(self.nonce),
            rlp.i2be_rlp_padded(self.gasPrice),
            rlp.i2be_rlp_padded(self.gas),
            bytes.fromhex(self.to[2:]),
            rlp.i2be_rlp_padded(self.value),
            self.data,
            rlp.i2be_rlp_padded(self.v),
            rlp.i2be_rlp_padded(self.r),
            rlp.i2be_rlp_padded(self.s)
        ]
        return rlp.encode(to_serialize)

    def _set_tx_id(self) -> None:
        '''Set the tx_id of the transaction object'''
        self.tx_id = crypto.keccak256(self.serialize())

    def sighash(self) -> bytes:
        '''
        Calculate the signature hash of the transaction. This is the digest
        that is signed by the sender.
        '''
        stripped = self.as_unsigned()
        return stripped.sighash()


class SignedCeloTx(CeloTx, SignedEthTx):
    '''
    A signed Ethereum transaction.

    Args:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        gasCurrency: the address of the token to be used as gas
        gasFeeRecipient: the address of the recipient of gas fees
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        v: EIP-155 signature v value
        r: signature r value
        s: signature s value

    Attributes:
        nonce: the tx nonce
        gasPrice: the number of wei to pay per unit of gas
        gas: the number of units of gas
        gasCurrency: the address of the token to be used as gas
        gasFeeRecipient: the address of the recipient of gas fees
        to: the address the tx is address to
        value: the number of wei attached to the tx
        data: bytes to be included in the tx's calldata
        v: EIP-155 signature v value
        r: signature r value
        s: signature s value
    '''
    def __init__(
            self,
            nonce: int,
            gasPrice: int,
            gas: int,
            gasCurrency: Optional[str],
            gasFeeRecipient: Optional[str],
            to: str,
            value: int,
            data: bytes,
            v: int,
            r: int,
            s: int):
        self.nonce = nonce
        self.gasPrice = gasPrice
        self.gas = gas
        self.gasCurrency = gasCurrency
        self.gasFeeRecipient = gasFeeRecipient
        self.to = to
        self.value = value
        self.data = data
        self.v = v
        self.r = r
        self.s = s
        self._set_tx_id()
        self._make_immutable()

    def as_unsigned(self) -> 'UnsignedCeloTx':
        '''Return an unsigned copy of the transaction'''
        # if r == 0 it's an unsigned tx anyway
        chainId = self.v if self.r == 0 \
            else (self.v - (35 if self.v % 2 else 36)) // 2
        return UnsignedCeloTx(
            nonce=self.nonce,
            gasPrice=self.gasPrice,
            gas=self.gas,
            gasCurrency=self.gasCurrency,
            gasFeeRecipient=self.gasFeeRecipient,
            to=self.to,
            value=self.value,
            data=self.data,
            chainId=chainId)

    @classmethod
    def deserialize(cls, raw: bytes) -> 'SignedCeloTx':
        '''Deserialize a transaction from raw bytes'''
        raws = cast(List[bytes], rlp.decode(raw))
        nonce = rlp.be2i_rlp(raws[0])
        gasPrice = rlp.be2i_rlp(raws[1])
        gas = rlp.be2i_rlp(raws[2])
        gasCurrency = None if raws[3] == b'' else f'0x{raws[3].hex()}'
        gasFeeRecipient = None if raws[4] == b'' else f'0x{raws[4].hex()}'
        to = f'0x{raws[3].hex()}'
        value = rlp.be2i_rlp(raws[4])
        data = raws[5]
        (v, r, s) = [rlp.be2i_rlp(i) for i in raws[6:]]

        return SignedCeloTx(
            nonce,
            gasPrice,
            gas,
            gasCurrency,
            gasFeeRecipient,
            to,
            value,
            data,
            v, r, s)

    def serialize(self) -> bytes:
        '''Serialize the transaction to raw bytes.'''
        gasCurrency = b''
        if self.gasCurrency is not None:
            gasCurrency = bytes.fromhex(self.gasCurrency[2:])

        gasFeeRecipient = b''
        if self.gasFeeRecipient is not None:
            gasFeeRecipient = bytes.fromhex(self.gasFeeRecipient[2:])

        to_serialize: List[bytes] = [
            rlp.i2be_rlp_padded(self.nonce),
            rlp.i2be_rlp_padded(self.gasPrice),
            rlp.i2be_rlp_padded(self.gas),
            gasCurrency,
            gasFeeRecipient,
            bytes.fromhex(self.to[2:]),
            rlp.i2be_rlp_padded(self.value),
            self.data,
            rlp.i2be_rlp_padded(self.v),
            rlp.i2be_rlp_padded(self.r),
            rlp.i2be_rlp_padded(self.s)
        ]
        return rlp.encode(to_serialize)



import abc
import ssl
import json
import aiohttp
import asyncio
import websockets

from ether.transactions import EthTx, SignedEthTx

from logging import Logger
from asyncio import Future
from websockets.client import WebSocketClientProtocol
from ether.ether_types import Receipt, RPCRequest
from ether.ether_types import RPCSubscription, UnparsedEtherEvent
from typing import Any, cast, Dict, Generator, List, Optional, Tuple, Union


def _id(start: int) -> Generator[int, None, None]:
    '''Infinite generator for unique request IDs'''
    index = start
    while True:
        yield index
        index += 1


def get_client(
        network: str,
        **kwargs: Any) -> 'BaseRPC':
    '''
    Convenience method for getting a client.
    Prefer instantiating directly

    Args:
        network (str): mainnet, ropsten, rinkeby, etc.
        force_https (bool): set true to disable websockets
    '''
    use_infura = 'infura_key' in kwargs \
                 and kwargs['infura_key'] is not None \
                 and len(kwargs['infura_key']) > 0
    force_https = 'force_https' in kwargs and kwargs['force_https']
    if use_infura:
        return InfuraWSRPC(network=network, **kwargs)
    if use_infura and force_https:
        return InfuraHTTPRPC(network=network, **kwargs)

    uri = ''
    if 'uri' in kwargs:
        uri = kwargs.pop('uri')
    else:
        raise ValueError('Must specify URI or infura key')
    if uri[0:3] == 'wss' and force_https:
        return WSRPC(uri, network, **kwargs)
    return HTTPRPC(uri, network, **kwargs)


class BaseRPC(metaclass=abc.ABCMeta):

    _ids: Generator[int, None, None]
    uri: str
    network: str
    connected: bool

    params: Dict[str, Any]

    def __repr__(self) -> str:
        return f'Eth RPC {self.mode} client: {self.network} via {self.uri}'

    def __init__(
            self,
            uri: str,
            network: str,
            logger: Optional[Logger] = None,
            **kwargs: Any) -> None:

        self._ids = _id(kwargs['_id'] if '_id' in kwargs else 0)
        self.connected = False

        self.uri = uri
        self.network = network
        self.mode = uri[0:3]
        self.params = kwargs

        self._logger = logger

    def info(self, *args: Any, **kwargs: Any) -> None:
        if self._logger is not None:
            self._logger.info(*args, **kwargs)

    def debug(self, *args: Any, **kwargs: Any) -> None:
        if self._logger is not None:
            self._logger.debug(*args, **kwargs)

    def warn(self, *args: Any, **kwargs: Any) -> None:
        if self._logger is not None:
            self._logger.warn(*args, **kwargs)

    def error(self, *args: Any, **kwargs: Any) -> None:
        if self._logger is not None:
            self._logger.error(*args, **kwargs)

    @abc.abstractmethod
    async def open(self) -> None:
        ...

    @abc.abstractmethod
    async def close(self) -> None:
        ...

    @abc.abstractmethod
    async def _RPC(self, method: str, params: List) -> Any:
        ...

    @staticmethod
    def _encode_if_int(v: Any) -> Any:
        if isinstance(v, int):
            return hex(cast(int, v))
        return v

    @staticmethod
    def _shallow_prep_params(params: List[Any]) -> List[Any]:
        return list(map(BaseRPC._encode_if_int, params))

    async def send_transaction(
            self,
            from_addr: str,
            tx: EthTx) -> str:
        '''Send a transaction, let the node sign it if it wants'''
        if hasattr(self, 'infura_key'):
            raise RuntimeError('Tried to sign tx with infura connection')

        param = cast(dict, tx.to_json_dict())
        param['from'] = from_addr

        for key, value in param.items():
            param[key] = BaseRPC._encode_if_int(value)

        res = await self._RPC(
            method='eth_sendTransaction',
            params=[param]
        )

        return cast(str, res)

    async def get_balance(
            self,
            address: str,
            block: Union[str, int] = 'latest') -> int:
        '''Gets the number of wei at an address'''
        res = await self._RPC(
            'eth_getBalance',
            [address, BaseRPC._encode_if_int(block)])
        return int(res, 16)

    async def get_logs(
            self,
            address: Optional[str] = None,
            from_block: Union[str, int] = 'earliest',
            to_block: Union[str, int] = 'latest',
            topics: Optional[List[Optional[str]]] = None,
            blockhash: Optional[str] = None) -> List[UnparsedEtherEvent]:
        '''Gets logs'''
        params: Dict[str, Union[str, List[Optional[str]]]] = {}

        if address:
            params['address'] = address
        if topics:
            params['topics'] = topics

        # can't have both a hash and a height range
        if blockhash:
            params['blockhash'] = blockhash
        else:
            params['fromBlock'] = BaseRPC._encode_if_int(from_block)
            params['toBlock'] = BaseRPC._encode_if_int(to_block)

        events = await self._RPC(method='eth_getLogs', params=[params])

        return cast(List[UnparsedEtherEvent], events)

    async def get_past_contract_logs(
            self,
            address: str,
            topics: Optional[List[Optional[str]]]) -> List[UnparsedEtherEvent]:
        '''Simpler method to get contract logs'''
        return await self.get_logs(address=address, topics=topics)

    async def broadcast(self, tx: str) -> str:
        '''Broadcast a transaction to the network'''
        if tx[0:2] != '0x':
            tx = f'0x{tx}'
        result = await self._RPC(
            method='eth_sendRawTransaction',
            params=[tx])
        return cast(str, result)

    async def get_tx_receipt(
            self,
            tx_id: str) -> Optional[Receipt]:
        '''Get the receipt of a transaction. None if not confirmed'''
        if tx_id[:2] != '0x':
            tx_id = f'0x{tx_id}'

        result = await self._RPC(
            method='eth_getTransactionReceipt',
            params=[tx_id])

        return cast(Optional[Receipt], result)

    async def get_nonce(self, account: str) -> int:
        res = await self._RPC(
            method='eth_getTransactionCount',
            params=[account, 'pending'])
        appropriate_nonce = int(res, 16)
        return appropriate_nonce

    async def preflight_tx(
            self,
            tx: EthTx,
            sender: Optional[str] = None) -> str:
        '''Preflight a transaction'''
        if sender is None:
            try:
                # if it doesn't have that name
                sender = cast(SignedEthTx, tx).recover_sender()
            except AttributeError:
                sender = '0x' + '11' * 20

        res = await self._RPC(
            method='eth_call',
            params=[
                {
                    'from': sender,
                    'to': tx.to,
                    'data': f'0x{tx.data.hex()}'
                },
                'latest'  # block height parameter
            ])
        return cast(str, res)


class WSRPC(BaseRPC):

    _ws: WebSocketClientProtocol

    _inflight: Dict[int, RPCRequest] = {}  # requests awaiting responses
    _subscriptions: Dict[str, RPCSubscription] = {}  # subscription queues

    _ping_task: Future
    _handle_task: Future

    def __init__(
            self,
            uri: str,
            network: str,
            **kwargs: Any) -> None:
        super().__init__(uri, network, **kwargs)

    def __repr__(self) -> str:
        '''Override repr to add connection status'''
        return f'{super().__repr__()}\tConnected: {self.connected}'

    async def open(self) -> None:
        self._ws = await websockets.connect(self.uri, ssl=ssl.SSLContext())
        self.connected = True
        self._ping_task = asyncio.ensure_future(self._ping())
        self._handle_task = asyncio.ensure_future(self._handle_incoming())

    async def _handle_incoming(self) -> None:
        '''Router for incoming messages. Distributes them to futures/queues'''
        try:
            while True:
                msg = await self._ws.recv()
                payload = json.loads(msg)
                # if it has a request ID, it's an RPC response
                if 'error' in payload:
                    self._inflight[payload['id']]['fut'].set_exception(
                        RuntimeError(str(payload['error'])))
                    continue
                elif 'id' in payload and payload['id'] in self._inflight:
                    future = self._inflight[payload['id']]['fut']
                    future.set_result(payload['result'])
                # otherwise it's a subscript notification
                # add it to the subscription's queue
                elif ('params' in payload
                      and 'subscription' in payload['params']):
                    sub_id = payload['params']['subscription']
                    q = self._subscriptions[sub_id]['queue']
                    await q.put(payload['params']['result'])
        except websockets.exceptions.ConnectionClosed:
            if self.connected:
                # if the forloop breaks due to a connection closed error
                await self.close()

        except asyncio.CancelledError:
            pass

    async def close(self) -> None:
        '''Close the socket and cancel associated tasks'''
        if not hasattr(self, '_ws'):
            return

        if not self.connected:
            return

        await self._ws.close()
        self.connected = False

        # prevent error loops
        if not self._ping_task.cancelled():
            self._ping_task.cancel()
            await self._ping_task

        # prevent error loops
        if not self._handle_task.cancel():
            self._handle_task.cancel()
            await self._handle_task

    def get_pending(
            self) -> Tuple[int, List[RPCRequest], List[RPCSubscription]]:
        if self.connected:
            raise RuntimeError(
                'Tried to retrive pending requests from active socket')

        pending: List[RPCRequest] = []
        subs: List[RPCSubscription] = []

        for _, r in self._inflight.items():
            if not r['fut'].done():
                pending.append(r)
        for _, s in self._subscriptions.items():
            subs.append(s)

        # what's a safe ID to resume from to prevent collisions
        next_id = max(self._inflight.keys()) + 1

        return next_id, pending, subs

    @classmethod
    async def from_pending(
            C,
            uri: str,
            network: str,
            start_id: int,
            pending: List[RPCRequest] = [],
            subs: List[RPCSubscription] = [],
            **kwargs: Any) -> 'WSRPC':
        '''
        Returns an open connection, using requests and subscriptions from
        another WSRPC instance. Subscriptions are restarted, and inflight reqs
        are remade. If you pass a pending request from another closed socket,
        the new socket will handle resolving the associated futures.

        Useful for seamlessly(ish) resuming a faulted session
        '''
        _id = start_id
        new = C(uri=uri, network=network, _id=_id, **kwargs)

        await new.open()

        reqs = [new._RPC(r['method'], r['params'], r['fut']) for r in pending]
        asyncio.ensure_future(asyncio.gather(*reqs))

        new_subs = [new._subscribe(s['params'], s['queue']) for s in subs]
        asyncio.ensure_future(asyncio.gather(*new_subs))
        return new

    async def _ping(self) -> None:
        try:
            while True:
                await self._ws.ping()
                await asyncio.sleep(15)
        except websockets.exceptions.ConnectionClosed:
            if self.connected:
                self.close()

    async def _RPC(
            self,
            method: str,
            params: List,
            fut: Optional[Future] = None) -> Any:
        '''Internal method for handling RPC calls'''
        # get a new ID and the websket
        req_id = next(self._ids)

        # Create a new future store it in our pendings and send it
        if fut is None:
            future = asyncio.get_event_loop().create_future()
        else:
            future = cast(Future, fut)
        self._inflight[req_id] = RPCRequest(
            method=method,
            params=params,
            fut=future)

        # send the request
        payload = {
            'jsonrpc': '2.0',
            'id': req_id,
            'method': method,
            'params': params
        }

        self.debug(f'dispatching ws payload {payload}')
        try:
            await self._ws.send(json.dumps(payload))
        except websockets.exceptions.ConnectionClosed:
            if self.connected:
                await self.close()
            raise RuntimeError('Websocket connection closed unexpectedly')

        # await the result
        res = await future
        return res

    async def _subscribe(
            self,
            params: List,
            queue: Optional[asyncio.Queue] = None) \
            -> Tuple[str, asyncio.Queue]:
        '''Internal method for setting up subscriptions'''
        method = 'eth_subscribe'

        # set up the queue
        q: asyncio.Queue
        if queue is None:
            q = asyncio.Queue()
        else:
            q = cast(asyncio.Queue, queue)

        # get the subscription ID and store the q
        sub_id = await self._RPC(method, params)
        self._subscriptions[sub_id] = RPCSubscription(
            method=method,
            params=params,
            queue=q)
        return sub_id, q

    async def unsubscribe(self, sub_ids: List[str]) -> bool:
        '''Cancel a list of subscriptions'''

        res = await self._RPC('eth_unsubscribe', sub_ids)
        for sub in sub_ids:
            self._subscriptions.pop(sub)
        return cast(bool, res)

    async def subscribe_to_address_events(
            self,
            addresses: List[str],
            topics: Optional[List[str]]) -> Tuple[str, asyncio.Queue]:
        '''Subscribes to event logs at specific addresses'''
        param_dict = {'address': addresses}
        if topics:
            param_dict['topics'] = topics

        return await self._subscribe(['logs', param_dict])

    async def subscribe_to_headers(self) -> Tuple[str, asyncio.Queue]:
        '''Subscribes to new headers on the network'''
        return await self._subscribe(['newHeads'])


class HTTPRPC(BaseRPC):

    _session: aiohttp.ClientSession

    def __init__(
            self,
            uri: str,
            network: str,
            **kwargs: Any) -> None:
        super().__init__(uri, network, **kwargs)

    async def open(self) -> None:
        if self.connected:
            return
        # close the connection after each request
        # some overhead, but less likely to get server-side disconnects
        self._session = aiohttp.ClientSession(
            headers={"Connection": "close"}
        )
        self.connected = True

    async def close(self) -> None:
        await self._session.close()
        self.connected = False

    async def _RPC(self, method: str, params: List[Any]) -> Any:
        payload = {
            "jsonrpc": "2.0",
            "id": next(self._ids),
            "method": method,
            "params": BaseRPC._shallow_prep_params(params)}
        self.debug(f'dispatching post request {payload}')
        resp = await self._session.post(self.uri, json=payload)
        if resp.status != 200:
            raise RuntimeError(f'Bad status during RPC request: {resp.status}')
        resp_json = await resp.json()
        result = resp_json['result'] if 'result' in resp_json else resp_json
        return result


class InfuraHTTPRPC(HTTPRPC):

    def __init__(self, network: str, infura_key: str, **kwargs: Any) -> None:
        uri = f'https://{network}.infura.io/v3/{infura_key}'
        kwargs['uri'] = uri
        self.infura_key = infura_key
        super().__init__(network=network, infura_key=infura_key, **kwargs)


class InfuraWSRPC(WSRPC):
    def __init__(self, network: str, infura_key: str, **kwargs: Any) -> None:
        uri = f'wss://{network}.infura.io/ws/v3/{infura_key}'
        kwargs['uri'] = uri
        self.infura_key = infura_key
        super().__init__(network=network, infura_key=infura_key, **kwargs)
